@@ -7,9 +7,20 @@ use crate::unwind_rule::UnwindRule;
 #[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub enum UnwindRuleAarch64 {
     /// (sp, fp, lr) = (sp, fp, lr)
+    /// Only possible for the first frame. Subsequent frames must get the
+    /// return address from somewhere other than the lr register to avoid
+    /// infinite loops.
     NoOp,
     /// (sp, fp, lr) = (sp + 16x, fp, lr)
+    /// Only possible for the first frame. Subsequent frames must get the
+    /// return address from somewhere other than the lr register to avoid
+    /// infinite loops.
     OffsetSp { sp_offset_by_16: u16 },
+    /// (sp, fp, lr) = (sp + 16x, fp, lr) if is_first_frame
+    /// This rule reflects an ambiguity in DWARF CFI information. When the
+    /// return address is "undefined" because it was omitted, it could mean
+    /// "same value", but this is only allowed for the first frame.
+    OffsetSpIfFirstFrameOtherwiseStackEndsHere { sp_offset_by_16: u16 },
     /// (sp, fp, lr) = (sp + 16x, fp, *(sp + 8y))
     OffsetSpAndRestoreLr {
         sp_offset_by_16: u16,
@@ -44,23 +55,46 @@ impl UnwindRule for UnwindRuleAarch64 {
         UnwindRuleAarch64::UseFramePointer
     }
 
-    fn exec<F>(self, regs: &mut UnwindRegsAarch64, read_stack: &mut F) -> Result<Option<u64>, Error>
+    fn exec<F>(
+        self,
+        is_first_frame: bool,
+        regs: &mut UnwindRegsAarch64,
+        read_stack: &mut F,
+    ) -> Result<Option<u64>, Error>
     where
         F: FnMut(u64) -> Result<u64, ()>,
     {
-        match self {
-            UnwindRuleAarch64::NoOp => {}
-            UnwindRuleAarch64::OffsetSp { sp_offset_by_16 } => {
-                let sp = regs.sp();
+        let lr = regs.lr();
+        let sp = regs.sp();
+        let fp = regs.fp();
+
+        let (new_lr, new_sp, new_fp) = match self {
+            UnwindRuleAarch64::NoOp => {
+                if !is_first_frame {
+                    return Err(Error::DidNotAdvance);
+                }
+                (lr, sp, fp)
+            }
+            UnwindRuleAarch64::OffsetSpIfFirstFrameOtherwiseStackEndsHere { sp_offset_by_16 } => {
+                if !is_first_frame {
+                    return Ok(None);
+                }
                 let sp_offset = u64::from(sp_offset_by_16) * 16;
                 let new_sp = sp.checked_add(sp_offset).ok_or(Error::IntegerOverflow)?;
-                regs.set_sp(new_sp);
+                (lr, new_sp, fp)
+            }
+            UnwindRuleAarch64::OffsetSp { sp_offset_by_16 } => {
+                if !is_first_frame {
+                    return Err(Error::DidNotAdvance);
+                }
+                let sp_offset = u64::from(sp_offset_by_16) * 16;
+                let new_sp = sp.checked_add(sp_offset).ok_or(Error::IntegerOverflow)?;
+                (lr, new_sp, fp)
             }
             UnwindRuleAarch64::OffsetSpAndRestoreLr {
                 sp_offset_by_16,
                 lr_storage_offset_from_sp_by_8,
             } => {
-                let sp = regs.sp();
                 let sp_offset = u64::from(sp_offset_by_16) * 16;
                 let new_sp = sp.checked_add(sp_offset).ok_or(Error::IntegerOverflow)?;
                 let lr_storage_offset = i64::from(lr_storage_offset_from_sp_by_8) * 8;
@@ -68,15 +102,13 @@ impl UnwindRule for UnwindRuleAarch64 {
                     checked_add_signed(sp, lr_storage_offset).ok_or(Error::IntegerOverflow)?;
                 let new_lr =
                     read_stack(lr_location).map_err(|_| Error::CouldNotReadStack(lr_location))?;
-                regs.set_sp(new_sp);
-                regs.set_lr(new_lr);
+                (new_lr, new_sp, fp)
             }
             UnwindRuleAarch64::OffsetSpAndRestoreFpAndLr {
                 sp_offset_by_16,
                 fp_storage_offset_from_sp_by_8,
                 lr_storage_offset_from_sp_by_8,
             } => {
-                let sp = regs.sp();
                 let sp_offset = u64::from(sp_offset_by_16) * 16;
                 let new_sp = sp.checked_add(sp_offset).ok_or(Error::IntegerOverflow)?;
                 let lr_storage_offset = i64::from(lr_storage_offset_from_sp_by_8) * 8;
@@ -89,9 +121,7 @@ impl UnwindRule for UnwindRuleAarch64 {
                     checked_add_signed(sp, fp_storage_offset).ok_or(Error::IntegerOverflow)?;
                 let new_fp =
                     read_stack(fp_location).map_err(|_| Error::CouldNotReadStack(fp_location))?;
-                regs.set_sp(new_sp);
-                regs.set_fp(new_fp);
-                regs.set_lr(new_lr);
+                (new_lr, new_sp, new_fp)
             }
             UnwindRuleAarch64::UseFramePointer => {
                 // Do a frame pointer stack walk. Frame-based aarch64 functions store the caller's fp and lr
@@ -132,7 +162,6 @@ impl UnwindRule for UnwindRuleAarch64 {
                 //  ^ sp                    ^ fp
                 //
                 // So: *fp is the caller's frame pointer, and *(fp + 8) is the return address.
-                let sp = regs.sp();
                 let fp = regs.fp();
                 let new_sp = fp.checked_add(16).ok_or(Error::IntegerOverflow)?;
                 let new_lr = read_stack(fp + 8).map_err(|_| Error::CouldNotReadStack(fp + 8))?;
@@ -143,18 +172,13 @@ impl UnwindRule for UnwindRuleAarch64 {
                 if new_fp <= fp || new_sp <= sp {
                     return Err(Error::FramepointerUnwindingMovedBackwards);
                 }
-                regs.set_sp(new_sp);
-                regs.set_fp(new_fp);
-                regs.set_lr(new_lr);
+                (new_lr, new_sp, new_fp)
             }
             UnwindRuleAarch64::UseFramepointerWithOffsets {
                 sp_offset_from_fp_by_8,
                 fp_storage_offset_from_fp_by_8,
                 lr_storage_offset_from_fp_by_8,
             } => {
-                let sp = regs.sp();
-                let fp = regs.fp();
-
                 let sp_offset_from_fp = u64::from(sp_offset_from_fp_by_8) * 8;
                 let new_sp = fp
                     .checked_add(sp_offset_from_fp)
@@ -176,13 +200,21 @@ impl UnwindRule for UnwindRuleAarch64 {
                 if new_fp <= fp || new_sp <= sp {
                     return Err(Error::FramepointerUnwindingMovedBackwards);
                 }
-                regs.set_sp(new_sp);
-                regs.set_fp(new_fp);
-                regs.set_lr(new_lr);
+                (new_lr, new_sp, new_fp)
             }
+        };
+        let return_address = regs.lr_mask().strip_ptr_auth(new_lr);
+        if return_address == 0 {
+            return Ok(None);
         }
+        if !is_first_frame && new_sp == sp {
+            return Err(Error::DidNotAdvance);
+        }
+        regs.set_lr(new_lr);
+        regs.set_sp(new_sp);
+        regs.set_fp(new_fp);
 
-        Ok(Some(regs.lr()))
+        Ok(Some(return_address))
     }
 }
 
@@ -197,18 +229,18 @@ mod test {
         ];
         let mut read_stack = |addr| Ok(stack[(addr / 8) as usize]);
         let mut regs = UnwindRegsAarch64::new(0x100300, 0x10, 0x20);
-        let res = UnwindRuleAarch64::NoOp.exec(&mut regs, &mut read_stack);
+        let res = UnwindRuleAarch64::NoOp.exec(true, &mut regs, &mut read_stack);
         assert_eq!(res, Ok(Some(0x100300)));
         assert_eq!(regs.sp(), 0x10);
-        let res = UnwindRuleAarch64::UseFramePointer.exec(&mut regs, &mut read_stack);
+        let res = UnwindRuleAarch64::UseFramePointer.exec(false, &mut regs, &mut read_stack);
         assert_eq!(res, Ok(Some(0x100200)));
         assert_eq!(regs.sp(), 0x30);
         assert_eq!(regs.fp(), 0x40);
-        let res = UnwindRuleAarch64::UseFramePointer.exec(&mut regs, &mut read_stack);
+        let res = UnwindRuleAarch64::UseFramePointer.exec(false, &mut regs, &mut read_stack);
         assert_eq!(res, Ok(Some(0x100100)));
         assert_eq!(regs.sp(), 0x50);
         assert_eq!(regs.fp(), 0x70);
-        let res = UnwindRuleAarch64::UseFramePointer.exec(&mut regs, &mut read_stack);
+        let res = UnwindRuleAarch64::UseFramePointer.exec(false, &mut regs, &mut read_stack);
         assert_eq!(res, Ok(None));
     }
 }
