@@ -150,7 +150,7 @@ pub struct UnwinderInternal<
     A: Arch + DwarfUnwinding + CompactUnwindInfoUnwinding + InstructionAnalysis,
     P: AllocationPolicy<D>,
 > {
-    /// sorted by address_range.start
+    /// sorted by avma_range.start
     modules: Vec<Module<D>>,
     /// Incremented every time modules is changed.
     modules_generation: u16,
@@ -187,13 +187,12 @@ impl<
     pub fn add_module(&mut self, module: Module<D>) {
         let insertion_index = match self
             .modules
-            .binary_search_by_key(&module.address_range.start, |module| {
-                module.address_range.start
-            }) {
+            .binary_search_by_key(&module.avma_range.start, |module| module.avma_range.start)
+        {
             Ok(i) => {
                 eprintln!(
                     "Now we have two modules at the same start address 0x{:x}. This can't be good.",
-                    module.address_range.start
+                    module.avma_range.start
                 );
                 i
             }
@@ -207,7 +206,7 @@ impl<
         if let Ok(index) = self
             .modules
             .binary_search_by_key(&module_address_range_start, |module| {
-                module.address_range.start
+                module.avma_range.start
             })
         {
             self.modules.remove(index);
@@ -216,13 +215,13 @@ impl<
     }
 
     pub fn max_known_code_address(&self) -> u64 {
-        self.modules.last().map_or(0, |m| m.address_range.end)
+        self.modules.last().map_or(0, |m| m.avma_range.end)
     }
 
     fn find_module_for_address(&self, address: u64) -> Option<(usize, u32)> {
         let (module_index, module) = match self
             .modules
-            .binary_search_by_key(&address, |m| m.address_range.start)
+            .binary_search_by_key(&address, |m| m.avma_range.start)
         {
             Ok(i) => (i, &self.modules[i]),
             Err(insertion_index) => {
@@ -232,18 +231,18 @@ impl<
                 }
                 let i = insertion_index - 1;
                 let module = &self.modules[i];
-                if module.address_range.end <= address {
+                if module.avma_range.end <= address {
                     // address is after this module
                     return None;
                 }
                 (i, module)
             }
         };
-        if address < module.base_address {
+        if address < module.base_avma {
             // Invalid base address
             return None;
         }
-        let relative_address = u32::try_from(address - module.base_address).ok()?;
+        let relative_address = u32::try_from(address - module.base_avma).ok()?;
         Some((module_index, relative_address))
     }
 
@@ -267,13 +266,14 @@ impl<
         ) -> Result<UnwindResult<A::UnwindRule>, UnwinderError>,
     {
         let lookup_address = address.address_for_lookup();
-        let cache_handle = match cache.rule_cache.try_unwind(
-            lookup_address,
-            self.modules_generation,
-            regs,
-            read_stack,
-        ) {
-            CacheResult::Hit(result) => return result,
+        let is_first_frame = !address.is_return_address();
+        let cache_handle = match cache
+            .rule_cache
+            .lookup(lookup_address, self.modules_generation)
+        {
+            CacheResult::Hit(unwind_rule) => {
+                return unwind_rule.exec(is_first_frame, regs, read_stack);
+            }
             CacheResult::Miss(handle) => handle,
         };
 
@@ -301,7 +301,7 @@ impl<
             }
         };
         cache.rule_cache.insert(cache_handle, unwind_rule);
-        unwind_rule.exec(regs, read_stack)
+        unwind_rule.exec(is_first_frame, regs, read_stack)
     }
 
     pub fn unwind_frame<F>(
@@ -328,28 +328,28 @@ impl<
     where
         F: FnMut(u64) -> Result<u64, ()>,
     {
+        let is_first_frame = !address.is_return_address();
         let unwind_result = match &module.unwind_data {
             ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(unwind_data, eh_frame_data) => {
                 // eprintln!("unwinding with cui and eh_frame in module {}", module.name);
                 let text_bytes = module.text_data.as_ref().and_then(|data| {
                     let offset_from_base =
-                        u32::try_from(data.address_range.start.checked_sub(module.base_address)?)
-                            .ok()?;
+                        u32::try_from(data.avma_range.start.checked_sub(module.base_avma)?).ok()?;
                     Some(TextBytes::new(offset_from_base, &data.bytes[..]))
                 });
-                let stubs_range = if let Some(stubs_range) = &module.sections.stubs {
+                let stubs_range = if let Some(stubs_range) = &module.svma_info.stubs {
                     (
-                        (stubs_range.start - module.base_address) as u32,
-                        (stubs_range.end - module.base_address) as u32,
+                        (stubs_range.start - module.svma_info.base_svma) as u32,
+                        (stubs_range.end - module.svma_info.base_svma) as u32,
                     )
                 } else {
                     (0, 0)
                 };
                 let stub_helper_range =
-                    if let Some(stub_helper_range) = &module.sections.stub_helper {
+                    if let Some(stub_helper_range) = &module.svma_info.stub_helper {
                         (
-                            (stub_helper_range.start - module.base_address) as u32,
-                            (stub_helper_range.end - module.base_address) as u32,
+                            (stub_helper_range.start - module.svma_info.base_svma) as u32,
+                            (stub_helper_range.end - module.svma_info.base_svma) as u32,
                         )
                     } else {
                         (0, 0)
@@ -360,7 +360,6 @@ impl<
                     stubs_range,
                     stub_helper_range,
                 );
-                let is_first_frame = !address.is_return_address();
 
                 let unwind_result = unwinder.unwind_frame(rel_lookup_address, is_first_frame)?;
                 match unwind_result {
@@ -375,10 +374,15 @@ impl<
                             UnwindSectionType::EhFrame,
                             None,
                             &mut cache.eh_frame_unwind_context,
-                            &module.sections,
+                            &module.svma_info,
                         );
-                        dwarf_unwinder
-                            .unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
+                        dwarf_unwinder.unwind_frame_with_fde(
+                            regs,
+                            is_first_frame,
+                            rel_lookup_address,
+                            fde_offset,
+                            read_stack,
+                        )?
                     }
                 }
             }
@@ -390,12 +394,18 @@ impl<
                     UnwindSectionType::EhFrame,
                     Some(eh_frame_hdr_data),
                     &mut cache.eh_frame_unwind_context,
-                    &module.sections,
+                    &module.svma_info,
                 );
                 let fde_offset = dwarf_unwinder
-                    .get_fde_offset_for_address(address.address_for_lookup())
+                    .get_fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::EhFrameHdrCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
+                dwarf_unwinder.unwind_frame_with_fde(
+                    regs,
+                    is_first_frame,
+                    rel_lookup_address,
+                    fde_offset,
+                    read_stack,
+                )?
             }
             ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, eh_frame_data) => {
                 let eh_frame_data = ArcData(eh_frame_data.clone());
@@ -404,12 +414,18 @@ impl<
                     UnwindSectionType::EhFrame,
                     None,
                     &mut cache.eh_frame_unwind_context,
-                    &module.sections,
+                    &module.svma_info,
                 );
                 let fde_offset = index
                     .fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::DwarfCfiIndexCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
+                dwarf_unwinder.unwind_frame_with_fde(
+                    regs,
+                    is_first_frame,
+                    rel_lookup_address,
+                    fde_offset,
+                    read_stack,
+                )?
             }
             ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(index, debug_frame_data) => {
                 let debug_frame_data = ArcData(debug_frame_data.clone());
@@ -418,12 +434,18 @@ impl<
                     UnwindSectionType::DebugFrame,
                     None,
                     &mut cache.eh_frame_unwind_context,
-                    &module.sections,
+                    &module.svma_info,
                 );
                 let fde_offset = index
                     .fde_offset_for_relative_address(rel_lookup_address)
                     .ok_or(UnwinderError::DwarfCfiIndexCouldNotFindAddress)?;
-                dwarf_unwinder.unwind_frame_with_fde(regs, address, fde_offset, read_stack)?
+                dwarf_unwinder.unwind_frame_with_fde(
+                    regs,
+                    is_first_frame,
+                    rel_lookup_address,
+                    fde_offset,
+                    read_stack,
+                )?
             }
             ModuleUnwindDataInternal::None => return Err(UnwinderError::NoModuleUnwindData),
         };
@@ -448,11 +470,7 @@ enum ModuleUnwindDataInternal<D: Deref<Target = [u8]>> {
 }
 
 impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
-    fn new(
-        unwind_data: ModuleUnwindData<D>,
-        sections: &ModuleSectionAddressRanges,
-        base_address: u64,
-    ) -> Self {
+    fn new(unwind_data: ModuleUnwindData<D>, svma_info: &ModuleSvmaInfo) -> Self {
         match unwind_data {
             ModuleUnwindData::CompactUnwindInfoAndEhFrame(cui, eh_frame) => {
                 ModuleUnwindDataInternal::CompactUnwindInfoAndEhFrame(cui, eh_frame.map(Arc::new))
@@ -461,7 +479,7 @@ impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
                 ModuleUnwindDataInternal::EhFrameHdrAndEhFrame(eh_frame_hdr, Arc::new(eh_frame))
             }
             ModuleUnwindData::EhFrame(eh_frame) => {
-                match DwarfCfiIndex::try_new_eh_frame(&eh_frame, sections, base_address) {
+                match DwarfCfiIndex::try_new_eh_frame(&eh_frame, svma_info) {
                     Ok(index) => {
                         ModuleUnwindDataInternal::DwarfCfiIndexAndEhFrame(index, Arc::new(eh_frame))
                     }
@@ -469,7 +487,7 @@ impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
                 }
             }
             ModuleUnwindData::DebugFrame(debug_frame) => {
-                match DwarfCfiIndex::try_new_debug_frame(&debug_frame, sections, base_address) {
+                match DwarfCfiIndex::try_new_debug_frame(&debug_frame, svma_info) {
                     Ok(index) => ModuleUnwindDataInternal::DwarfCfiIndexAndDebugFrame(
                         index,
                         Arc::new(debug_frame),
@@ -484,42 +502,43 @@ impl<D: Deref<Target = [u8]>> ModuleUnwindDataInternal<D> {
 
 pub struct TextByteData<D: Deref<Target = [u8]>> {
     bytes: D,
-    address_range: Range<u64>,
+    avma_range: Range<u64>,
 }
 
 impl<D: Deref<Target = [u8]>> TextByteData<D> {
-    pub fn new(bytes: D, address_range: Range<u64>) -> Self {
-        Self {
-            bytes,
-            address_range,
-        }
+    pub fn new(bytes: D, avma_range: Range<u64>) -> Self {
+        Self { bytes, avma_range }
     }
 
-    pub fn get_bytes(&self, address_range: Range<u64>) -> Option<&[u8]> {
-        let rel_start = address_range.start.checked_sub(self.address_range.start)?;
+    pub fn get_bytes(&self, avma_range: Range<u64>) -> Option<&[u8]> {
+        let rel_start = avma_range.start.checked_sub(self.avma_range.start)?;
         let rel_start = usize::try_from(rel_start).ok()?;
-        let rel_end = address_range.end.checked_sub(self.address_range.start)?;
+        let rel_end = avma_range.end.checked_sub(self.avma_range.start)?;
         let rel_end = usize::try_from(rel_end).ok()?;
         self.bytes.get(rel_start..rel_end)
     }
 
-    pub fn address_range(&self) -> Range<u64> {
-        self.address_range.clone()
+    pub fn avma_range(&self) -> Range<u64> {
+        self.avma_range.clone()
     }
 }
 
 pub struct Module<D: Deref<Target = [u8]>> {
     #[allow(unused)]
     name: String,
-    address_range: Range<u64>,
-    base_address: u64,
-    sections: ModuleSectionAddressRanges,
+    avma_range: Range<u64>,
+    base_avma: u64,
+    svma_info: ModuleSvmaInfo,
     unwind_data: ModuleUnwindDataInternal<D>,
     text_data: Option<TextByteData<D>>,
 }
 
+/// The addresses of various sections in the module.
+/// These are SVMAs, "stated virtual memory addresses", i.e. addresses as stated
+/// in the object.
 #[derive(Clone, Debug, PartialEq, Eq)]
-pub struct ModuleSectionAddressRanges {
+pub struct ModuleSvmaInfo {
+    pub base_svma: u64,
     pub text: Option<Range<u64>>,
     pub text_env: Option<Range<u64>>,
     pub stubs: Option<Range<u64>>,
@@ -532,18 +551,18 @@ pub struct ModuleSectionAddressRanges {
 impl<D: Deref<Target = [u8]>> Module<D> {
     pub fn new(
         name: String,
-        address_range: std::ops::Range<u64>,
-        base_address: u64,
-        sections: ModuleSectionAddressRanges,
+        avma_range: std::ops::Range<u64>,
+        base_avma: u64,
+        svma_info: ModuleSvmaInfo,
         unwind_data: ModuleUnwindData<D>,
         text_data: Option<TextByteData<D>>,
     ) -> Self {
-        let unwind_data = ModuleUnwindDataInternal::new(unwind_data, &sections, base_address);
+        let unwind_data = ModuleUnwindDataInternal::new(unwind_data, &svma_info);
         Self {
             name,
-            address_range,
-            base_address,
-            sections,
+            avma_range,
+            base_avma,
+            svma_info,
             unwind_data,
             text_data,
         }
